# ==================================================================================================
# Copyright 2012 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

from functools import reduce

from twitter.pants.tasks import Task, TaskError

from twitter.pants.python.interpreter_cache import PythonInterpreterCache

from .target import is_python_root


class SetupPythonEnvironment(Task):
  """
    Establishes the python intepreter(s) for downstream Python tasks e.g. Resolve, Run, PytestRun.

    Populates the product namespace (for typename = 'python'):
      'intepreters': ordered list of PythonInterpreter objects
  """
  @classmethod
  def setup_parser(cls, option_group, args, mkflag):
    option_group.add_option(mkflag("force"), dest="python_setup_force",
                            action="store_true", default=False,
                            help="Force clean and install.")
    option_group.add_option(mkflag("path"), dest="python_setup_paths",
                            action="append", default=[],
                            help="Add a path to search for interpreters, by default PATH.")
    option_group.add_option(mkflag("interpreter"), dest="python_interpreter",
                            default=[], action='append',
                            help="Constrain what Python interpreters to use.  Uses Requirement "
                                 "format from pkg_resources, e.g. 'CPython>=2.6,<3' or 'PyPy'. "
                                 "By default, no constraints are used.  Multiple constraints may "
                                 "be added.  They will be ORed together.")
    option_group.add_option(mkflag("multi"), dest="python_multi",
                            default=False, action='store_true',
                            help="Allow multiple interpreters to be bound to an upstream chroot.")

  def __init__(self, context):
    context.products.require('python')
    self._cache = PythonInterpreterCache(context.config, logger=context.log.debug)
    super(SetupPythonEnvironment, self).__init__(context)

  def execute(self, _):
    ifilters = self.context.options.python_interpreter
    self._cache.setup(force=self.context.options.python_setup_force,
        paths=self.context.options.python_setup_paths,
        filters=ifilters or [''])
    all_interpreters = set(self._cache.interpreters)
    for target in self.context.targets(is_python_root):
      self.context.log.info('Setting up interpreters for %s' % target)
      closure = target.closure()
      self.context.log.debug('  - Target closure: %d targets' % len(closure))
      target_compatibilities = [
          set(self._cache.matches(getattr(closure_target, 'compatibility', [''])))
          for closure_target in closure]
      target_compatibilities = reduce(set.intersection, target_compatibilities, all_interpreters)
      self.context.log.debug('  - Target minimum compatibility: %s' % (
        ' '.join(interp.version_string for interp in target_compatibilities)))
      interpreters = self._cache.select_interpreter(target_compatibilities,
          allow_multiple=self.context.options.python_multi)
      self.context.log.debug('  - Selected: %s' % interpreters)
      if not interpreters:
        raise TaskError('No compatible interpreters for %s' % target)
      target.interpreters = interpreters



# ==================================================================================================
# Copyright 2013 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

import subprocess
import errno


class Xargs(object):
  """A subprocess execution wrapper in the spirit of the xargs command line tool.

  Specifically allows encapsulated commands to be passed very large argument lists by chunking up
  the argument lists into a minimal set and then invoking the encapsulated command against each
  chunk in turn.
  """

  @classmethod
  def subprocess(cls, cmd, **kwargs):
    """Creates an xargs engine that uses subprocess.call to execute the given cmd array with extra
    arg chunks.
    """
    def call(args):
      return subprocess.call(cmd + args, **kwargs)
    return cls(call)

  def __init__(self, cmd):
    """Creates an xargs engine that calls cmd with argument chunks.

    :param cmd: A function that can execute a command line in the form of a list of strings
      passed as its sole argument.
    """
    self._cmd = cmd

  def _split_args(self, args):
    half = len(args) // 2
    return args[:half], args[half:]

  def execute(self, args):
    """Executes the configured cmd passing args in one or more rounds xargs style.

    :param list args: Extra arguments to pass to cmd.
    """
    all_args = list(args)
    try:
      return self._cmd(all_args)
    except OSError as e:
      if errno.E2BIG == e.errno:
        args1, args2 = self._split_args(all_args)
        result = self.execute(args1)
        if result != 0:
          return result
        return self.execute(args2)
      else:
        raise e



# ==================================================================================================
# Copyright 2012 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

import os

from functools import partial

from twitter.common.collections import maybe_list
from twitter.common.dirutil import Fileset
from twitter.common.lang import Compatibility
from twitter.pants import is_concrete
from twitter.pants.base import manual, TargetDefinitionException

from .internal import InternalTarget
from .jvm_target import JvmTarget
from .pants_target import Pants
from .resources import Resources


@manual.builddict(tags=["jvm"])
class JvmBinary(JvmTarget):
  """Produces a JVM binary optionally identifying a launcher main class.

  Below are a summary of how key goals affect targets of this type:

  * ``bundle`` - Creates a self-contained directory with the binary and all
    its dependencies, optionally archived, suitable for deployment.
  * ``binary`` - Create an executable jar of the binary. On the JVM
    this means the jar has a manifest specifying the main class.
  * ``run`` - Executes the main class of this binary locally.
  """
  def __init__(self, name,
               main=None,
               basename=None,
               source=None,
               resources=None,
               dependencies=None,
               excludes=None,
               deploy_excludes=None,
               configurations=None):
    """
    :param string name: The name of this target, which combined with this
      build file defines the target :class:`twitter.pants.base.address.Address`.
    :param string main: The name of the ``main`` class, e.g.,
      ``'com.twitter.common.examples.pingpong.Main'``. This class may be
      present as the source of this target or depended-upon library.
    :param string basename: Base name for the generated ``.jar`` file, e.g.,
      ``'pingpong'``. (By default, uses ``name`` param)
    :param string source: Name of one ``.java`` or ``.scala`` file (a good
      place for a ``main``).
    :param resources: List of ``resource``\s to include in bundle.
    :param dependencies: List of targets (probably ``java_library`` and
     ``scala_library`` targets) to "link" in.
    :param excludes: List of ``exclude``\s to filter this target's transitive
     dependencies against.
    :param deploy_excludes: List of ``excludes`` to apply at deploy time.
      If you, for example, deploy a java servlet that has one version of
      ``servlet.jar`` onto a Tomcat environment that provides another version,
      they might conflict. ``deploy_excludes`` gives you a way to build your
      code but exclude the conflicting ``jar`` when deploying.
    :param configurations: Ivy configurations to resolve for this target.
      This parameter is not intended for general use.
    :type configurations: tuple of strings
    """
    JvmTarget.__init__(self,
                       name=name,
                       sources=[source] if source else None,
                       dependencies=dependencies,
                       excludes=excludes,
                       configurations=configurations)

    if main and not isinstance(main, Compatibility.string):
      raise TargetDefinitionException(self, 'main must be a fully qualified classname')

    if source and not isinstance(source, Compatibility.string):
      raise TargetDefinitionException(self, 'source must be a single relative file path')

    self.main = main
    self.basename = basename or name
    self.resources = list(self.resolve_all(resources, Resources))
    self.deploy_excludes = deploy_excludes or []


class RelativeToMapper(object):
  """A mapper that maps files specified relative to a base directory."""

  def __init__(self, base):
    """The base directory files should be mapped from."""

    self.base = base

  def __call__(self, file):
    return os.path.relpath(file, self.base)

  def __repr__(self):
    return 'IdentityMapper(%s)' % self.base


@manual.builddict(tags=["jvm"])
class Bundle(object):
  """A set of files to include in an application bundle.

  Files added to the bundle will be included when bundling an application target.
  By default relative paths are preserved. For example, to include ``config``
  and ``scripts`` directories: ::

    bundles=[
      bundle().add(rglobs('config/*', 'scripts/*')),
    ]

  To include files relative to some path component use the ``relative_to`` parameter.
  The following places the contents of ``common/config`` in a  ``config`` directory
  in the bundle. ::

    bundles=[
      bundle(relative_to='common').add(globs('common/config/*'))
    ]
  """

  def __init__(self, mapper=None, relative_to=None):
    """
    :param mapper: Function that takes a path string and returns a path string. Takes a path in
      the source tree, returns a path to use in the resulting bundle. By default, an identity mapper.
    :param string relative_to: Set up a simple mapping from source path to bundle path.
      E.g., ``relative_to='common'`` removes that prefix from all files in the application bundle.
    """
    if mapper and relative_to:
      raise ValueError("Must specify exactly one of 'mapper' or 'relative_to'")

    if relative_to:
      base = os.path.abspath(relative_to)
      if not os.path.isdir(base):
        raise ValueError('Could not find a directory to bundle relative to at %s' % base)
      self.mapper = RelativeToMapper(base)
    else:
      self.mapper = mapper or RelativeToMapper(os.getcwd())

    self.filemap = {}

  @manual.builddict()
  def add(self, *filesets):
    """Add files to the bundle, where ``filesets`` is a filename, ``globs``, or ``rglobs``.
    Note this is a variable length param and may be specified any number of times.
    """
    for fileset in filesets:
      paths = fileset() if isinstance(fileset, Fileset) \
                        else fileset if hasattr(fileset, '__iter__') \
                        else [fileset]
      self.filemap.update(((os.path.abspath(path), self.mapper(path)) for path in paths))
    return self

  def resolve(self):
    yield self

  def __repr__(self):
    return 'Bundle(%s, %s)' % (self.mapper, self.filemap)


@manual.builddict(tags=["jvm"])
class JvmApp(InternalTarget):
  """A JVM-based application consisting of a binary plus "extra files".

  Invoking the ``bundle`` goal on one of these targets creates a
  self-contained artifact suitable for deployment on some other machine.
  The artifact contains the executable jar, its dependencies, and
  extra files like config files, startup scripts, etc.
  """

  def __init__(self, name, binary, bundles, basename=None):
    """
    :param string name: The name of this target, which combined with this
      build file defines the target :class:`twitter.pants.base.address.Address`.
    :param binary: The :class:`twitter.pants.targets.jvm_binary.JvmBinary`,
      or a :class:`twitter.pants.targets.pants_target.Pants` pointer to one.
    :param bundles: One or more :class:`twitter.pants.targets.jvm_binary.Bundle`'s
      describing "extra files" that should be included with this app
      (e.g.: config files, startup scripts).
    :param string basename: Name of this application, if different from the
      ``name``. Pants uses this in the ``bundle`` goal to name the distribution
      artifact. In most cases this parameter is not necessary.
    """
    InternalTarget.__init__(self, name, dependencies=[binary])

    binary_list = list()
    for b in maybe_list(binary, expected_type=(Pants, JvmBinary),
                          raise_type=partial(TargetDefinitionException, self)):
      binary_list.extend(b.resolve())
    if len(binary_list) != 1:
      raise TargetDefinitionException(
        self, 'binary must resolve to a JvmBinary target, found %s' % binary)

    self._binary = binary
    self._bundles = maybe_list(bundles, expected_type=(Bundle),
                               raise_type=partial(TargetDefinitionException, self))

    if name == basename:
      raise TargetDefinitionException(self, 'basename must not equal name.')
    self.basename = basename or name

    self._resolved_binary = None
    self._resolved_bundles = []

  @property
  def binary(self):
    self._maybe_resolve_binary()
    return self._resolved_binary

  def _maybe_resolve_binary(self):
    if self._binary is not None:
      binaries = list(filter(is_concrete, self._binary.resolve()))
      if len(binaries) != 1 or not isinstance(binaries[0], JvmBinary):
        raise TargetDefinitionException(self,
                                        'must supply exactly 1 JvmBinary, got %s' % self._binary)
      self._resolved_binary = binaries[0]
      self._binary = None

  @property
  def bundles(self):
    self._maybe_resolve_bundles()
    return self._resolved_bundles

  def _maybe_resolve_bundles(self):
    if self._bundles is not None:
      def is_resolvable(item):
        return hasattr(item, 'resolve')

      def is_bundle(bundle):
        return isinstance(bundle, Bundle)

      def resolve(item):
        return list(item.resolve()) if is_resolvable(item) else [None]

      if is_resolvable(self._bundles):
        self._bundles = resolve(self._bundles)

      try:
        for item in iter(self._bundles):
          for bundle in resolve(item):
            if not is_bundle(bundle):
              raise TypeError()
            self._resolved_bundles.append(bundle)
      except TypeError:
        raise TargetDefinitionException(self, 'bundles must be one or more Bundle objects, '
                                              'got %s' % self._bundles)
      self._bundles = None



# ==================================================================================================
# Copyright 2011 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

import os
import sys

from twitter.common.collections import maybe_list
from twitter.common.lang import Compatibility
from twitter.common.python.pex_info import PexInfo
from twitter.common.python.platforms import Platform

from twitter.pants.base import manual, TargetDefinitionException
from twitter.pants.base.build_info import get_build_info
from .python_target import PythonTarget


@manual.builddict(tags=["python"])
class PythonBinary(PythonTarget):
  """Produces a Python binary.

  Python binaries are pex files, self-contained executable shell
  scripts that contain a complete Python environment capable of
  running the target. For more information about pex files see
  https://github.com/twitter/commons/blob/master/src/python/twitter/pants/python/README.md"""

  # TODO(wickman) Consider splitting pex options out into a separate PexInfo builder that can be
  # attached to the binary target.  Ideally the PythonBinary target is agnostic about pex mechanics
  def __init__(self,
               name,
               source=None,
               dependencies=None,
               entry_point=None,
               inherit_path=False,        # pex option
               zip_safe=True,             # pex option
               always_write_cache=False,  # pex option
               repositories=None,         # pex option
               indices=None,              # pex option
               ignore_errors=False,       # pex option
               allow_pypi=False,          # pex option
               platforms=(),
               compatibility=None):
    """
    :param name: target name
    :param source: the python source file that becomes this binary's __main__.
      If None specified, drops into an interpreter by default.
    :param dependencies: List of :class:`twitter.pants.base.target.Target` instances
      this target depends on.
    :type dependencies: list of targets
    :param entry_point: the default entry point for this binary.  if None, drops into the entry
      point that is defined by source
    :param inherit_path: inherit the sys.path of the environment that this binary runs in
    :param zip_safe: whether or not this binary is safe to run in compacted (zip-file) form
    :param always_write_cache: whether or not the .deps cache of this PEX file should always
      be written to disk.
    :param repositories: a list of repositories to query for dependencies.
    :param indices: a list of indices to use for packages.
    :param platforms: extra platforms to target when building this binary.
    :param compatibility: either a string or list of strings that represents
      interpreter compatibility for this target, using the Requirement-style format,
      e.g. ``'CPython>=3', or just ['>=2.7','<3']`` for requirements agnostic to interpreter class.
    """
    if source is None and entry_point is None:
      raise TargetDefinitionException(self,
          'A python binary target must specify either source or entry_point.')

    PythonTarget.__init__(self,
        name,
        [] if source is None else [source],
        compatibility=compatibility,
        dependencies=dependencies,
    )

    if not isinstance(platforms, (list, tuple)) and not isinstance(platforms, Compatibility.string):
      raise TargetDefinitionException(self,
          'platforms must be a list, tuple or string.')

    self._entry_point = entry_point
    self._inherit_path = bool(inherit_path)
    self._zip_safe = bool(zip_safe)
    self._always_write_cache = bool(always_write_cache)
    self._repositories = maybe_list(repositories or [])
    self._indices = maybe_list(indices or [])
    self._ignore_errors = bool(ignore_errors)
    self._platforms = tuple(maybe_list(platforms or []))

    if source and entry_point:
      entry_point_module = entry_point.split(':', 1)[0]
      source_entry_point = self._translate_to_entry_point(self.sources[0])
      if entry_point_module != source_entry_point:
        raise TargetDefinitionException(self,
            'Specified both source and entry_point but they do not agree: %s vs %s' % (
            source_entry_point, entry_point_module))

  @property
  def platforms(self):
    return self._platforms

  # TODO(wickman) These should likely be attributes on PythonLibrary targets
  # and not PythonBinary targets, or at the very worst, both.
  @property
  def repositories(self):
    return self._repositories

  @property
  def indices(self):
    return self._indices

  def _translate_to_entry_point(self, source):
    source_base, _ = os.path.splitext(source)
    return source_base.replace(os.path.sep, '.')

  @property
  def entry_point(self):
    if self._entry_point:
      return self._entry_point
    elif self.sources:
      assert len(self.sources) == 1
      return self._translate_to_entry_point(self.sources[0])
    else:
      return None

  @property
  def pexinfo(self):
    info = PexInfo.default()
    info.build_properties = get_build_info()._asdict()
    for repo in self._repositories:
      info.add_repository(repo)
    for index in self._indices:
      info.add_index(index)
    info.zip_safe = self._zip_safe
    info.always_write_cache = self._always_write_cache
    info.inherit_path = self._inherit_path
    info.entry_point = self.entry_point
    info.ignore_errors = self._ignore_errors
    return info



# ==================================================================================================
# Copyright 2013 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

import os
import stat
import zipfile

from contextlib import contextmanager

from twitter.common.contextutil import open_zip

from .manifest import Manifest


@contextmanager
def open_jar(path, *args, **kwargs):
  """Yields a jar in a with context that will be closed when the context exits.

  The yielded jar is a zipfile.ZipFile object with an additional mkdirs(arcpath) method that will
  create a zip directory entry similar to unix `mkdir -p`.  Additionally, the ZipFile.write and
  ZipFile.writestr methods are enhanced to call mkdirs as needed to ensure all jar entries contain
  a full complement of parent paths leading from each leaf to the root of the jar.
  """

  with open_zip(path, *args, **kwargs) as jar:
    real_write = jar.write
    real_writestr = jar.writestr

    made_dirs = set()

    def mkdirs(arcpath):
      if arcpath and arcpath not in made_dirs:
        made_dirs.add(arcpath)

        parent_path = os.path.dirname(arcpath)
        mkdirs(parent_path)

        zipinfo = zipfile.ZipInfo(arcpath if arcpath.endswith('/') else arcpath + '/')

        # We store directories without compression since they have no contents and
        # attempts to store them with compression lead to corrupted zip files as such:
        # $ unzip -t junit-runner-0.0.19.jar
        # Archive:  junit-runner-0.0.19.jar
        # testing: com/
        # error:  invalid compressed data to inflate
        # testing: com/twitter/
        # error:  invalid compressed data to inflate
        # testing: com/twitter/common/
        # error:  invalid compressed data to inflate
        # testing: com/twitter/common/testing/
        # error:  invalid compressed data to inflate
        # testing: com/twitter/common/testing/runner/
        # error:  invalid compressed data to inflate
        # testing: com/twitter/common/testing/runner/StreamSource.class   OK
        zipinfo.compress_type = zipfile.ZIP_STORED

        # PKZIP says external_attr is a 4 byte field that is host system dependant:
        #   http://www.pkware.com/documents/casestudies/APPNOTE.TXT
        # These notes do mention the low order byte will carry DOS file attributes for DOS host
        # system zips.  The DOS file attribute bits are described here:
        #   http://www.xxcopy.com/xxcopy06.htm
        #
        # More details are only found reading source, for example in BSD:
        #   ftp://ftp-archive.freebsd.org/pub/FreeBSD-Archive/old-releases/i386/1.0-RELEASE/ports/info-zip/zipinfo/zipinfo.c
        # These sources reveal the 2 high order bytes contain unix file attribute bits.
        #
        # In summary though the full 32 bit field layout is:
        # TTTTsstrwxrwxrwx0000000000ADVSHR
        # ^^^^____________________________ stat.h file type: S_IFXXX
        #     ^^^_________________________ setuid, setgid, sticky
        #        ^^^^^^^^^________________ permissions
        #                 ^^^^^^^^________ ???
        #                         ^^^^^^^^ DOS attribute bits

        # Setup unix directory perm bits: drwxr-xr-x
        zipinfo.external_attr = (
          stat.S_IFDIR                                  # file type dir
          | stat.S_IRUSR | stat.S_IWUSR | stat.S_IXUSR  # u+rwx
          | stat.S_IRGRP | stat.S_IXGRP                 # g+rx
          | stat.S_IROTH | stat.S_IXOTH                 # o+rx
        ) << 16

        # Add DOS directory bit
        zipinfo.external_attr |= 0x10

        real_writestr(zipinfo, '')

    def write(path, arcname=None, **kwargs):
      if os.path.isdir(path):
        mkdirs(arcname or path)
      else:
        mkdirs(os.path.dirname(arcname or path))
        real_write(path, arcname, **kwargs)

    def writestr(zinfo_or_arcname, *args, **kwargs):
      mkdirs(os.path.dirname(zinfo_or_arcname))
      real_writestr(zinfo_or_arcname, *args, **kwargs)

    jar.mkdirs = mkdirs
    jar.write = write
    jar.writestr = writestr

    yield jar


__all__ = (
    'open_jar',
    'Manifest',
)



# ==================================================================================================
# Copyright 2013 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

import hashlib
import os
import shutil

from twitter.common.collections import maybe_list
from twitter.common.contextutil import temporary_file
from twitter.common.dirutil import safe_delete, touch
from twitter.common.lang import Compatibility
from twitter.common.quantity import Amount, Time

from twitter.pants.base.config import Config
from twitter.pants.java import Executor, SubprocessExecutor
from twitter.pants.net.http.fetcher import Fetcher


class Bootstrapper(object):
  """Bootstraps a working ivy resolver.

  By default a working resolver will be bootstrapped from maven central and it will use standard
  public jar repositories and a standard ivy local cache directory to execute resolve operations.

  A combination of site configuration options and environment variables can be used to override this
  default setup.

  By default ivy will be bootstrapped from a stable ivy jar version found in maven central, but
  this can be over-ridden with the ``ivy / bootstrap_jar_url`` config option.  Additionally the
  bootstrapping will use a connect/read timeout of 1 second by default, but this can be raised by
  specifying a ``ivy / bootstrap_fetch_timeout_secs`` config value.

  After bootstrapping, ivy will re-resolve itself.  By default it does this via maven central, but
  a custom ivy tool classpath can be specified by using the ``ivy / ivy_profile`` option to point to
  a custom ivy profile ivy.xml.  This can be useful to upgrade ivy to a version released after pants
  or else mix in auxiliary jars that provide ivy plugins.

  Finally, by default the ivysettings.xml embedded in the ivy jar will be used in conjunction with
  the default ivy local cache directory of ~/.ivy2/cache.  To specify custom values for these you
  can either provide ``ivy / ivy_settings`` and ``ivy / cache_dir`` config values or supply these
  values via the ``PANTS_IVY_SETTINGS_XML`` and ``PANTS_IVY_CACHE_DIR`` environment variables
  respectively.  The environment variables will trump config values if present.
  """

  class Error(Exception):
    """Indicates an error bootstrapping an ivy classpath."""

  _DEFAULT_VERSION = '2.3.0'
  _DEFAULT_URL = ('http://repo1.maven.org/maven2/'
                  'org/apache/ivy/ivy/'
                  '%(version)s/ivy-%(version)s.jar' % {'version': _DEFAULT_VERSION})

  _INSTANCE = None

  @classmethod
  def instance(cls):
    """Returns the default global ivy bootstrapper."""
    if cls._INSTANCE is None:
      cls._INSTANCE = cls()
    return cls._INSTANCE

  @classmethod
  def default_ivy(cls, java_executor=None):
    """Returns an Ivy instance using the default global bootstrapper.

    By default runs ivy via a subprocess java executor.

    :param java_executor: the optional java executor to use
    :returns: an Ivy instance.
    :raises: Bootstrapper.Error if the default ivy instance could not be bootstrapped
    """
    return cls.instance().ivy(java_executor=java_executor)

  def __init__(self, config=None):
    """Creates an ivy bootstrapper.

    :param config: the pants repo config to obtain ivy settings values from
    """
    self._config = config or Config.load()
    self._bootstrap_jar_url = self._config.get('ivy', 'bootstrap_jar_url',
                                               default=self._DEFAULT_URL)
    self._timeout = Amount(self._config.getint('ivy', 'bootstrap_fetch_timeout_secs', default=1),
                           Time.SECONDS)
    self._version_or_ivyxml = self._config.get('ivy', 'ivy_profile', default=self._DEFAULT_VERSION)
    self._classpath = None

  def ivy(self, java_executor=None):
    """Returns an ivy instance bootstrapped by this bootstrapper.

    :raises: Bootstrapper.Error if ivy could not be bootstrapped
    """
    return Ivy(self.classpath,
               java_executor=java_executor,
               ivy_settings=self.ivy_settings,
               ivy_cache_dir=self.ivy_cache_dir)

  @property
  def classpath(self):
    """Returns the bootstrapped ivy classpath as a list of jar paths.

    :raises: Bootstrapper.Error if the classpath could not be bootstrapped
    """
    if not self._classpath:
      self._classpath = self._bootstrap_ivy_classpath()
    return self._classpath

  @property
  def ivy_settings(self):
    """Returns the bootstrapped ivysettings.xml path.

    By default the ivy.ivy_settings value found in pants.ini but can be overridden by via the
    PANTS_IVY_SETTINGS_XML environment variable.  If neither is specified defaults to ivy's built
    in default ivysettings.xml of standard public resolvers.
    """
    return os.getenv('PANTS_IVY_SETTINGS_XML') or self._config.get('ivy', 'ivy_settings')

  @property
  def ivy_cache_dir(self):
    """Returns the bootstrapped ivy cache dir.

    By default the ivy.cache_dir value found in pants.ini but can be overridden via the
    PANTS_IVY_CACHE_DIR environment variable.  If neither is specified defaults to ivy's built
    in default cache dir; ie: ~/.ivy2/cache.
    """
    return (os.getenv('PANTS_IVY_CACHE_DIR')
            or self._config.get('ivy', 'cache_dir', default=os.path.expanduser('~/.ivy2/cache')))

  def _bootstrap_ivy_classpath(self, retry=True):
    # TODO(John Sirois): Extract a ToolCache class to control the path structure:
    # https://jira.twitter.biz/browse/DPB-283
    ivy_cache = os.path.join(self._config.getdefault('pants_cachedir'), 'tools', 'jvm', 'ivy')

    digest = hashlib.sha1()
    if os.path.isfile(self._version_or_ivyxml):
      with open(self._version_or_ivyxml) as fp:
        digest.update(fp.read())
    else:
      digest.update(self._version_or_ivyxml)
    classpath = os.path.join(ivy_cache, '%s.classpath' % digest.hexdigest())

    if not os.path.exists(classpath):
      ivy = self._bootstrap_ivy(os.path.join(ivy_cache, 'bootstrap.jar'))
      args = ['-confs', 'default', '-cachepath', classpath]
      if os.path.isfile(self._version_or_ivyxml):
        args.extend(['-ivy', self._version_or_ivyxml])
      else:
        args.extend(['-dependency', 'org.apache.ivy', 'ivy', self._version_or_ivyxml])

      try:
        ivy.execute(args)
      except ivy.Error as e:
        safe_delete(classpath)
        raise self.Error('Failed to bootstrap an ivy classpath! %s' % e)

    with open(classpath) as fp:
      cp = fp.read().strip().split(os.pathsep)
      if not all(map(os.path.exists, cp)):
        safe_delete(classpath)
        if retry:
          return self._bootstrap_ivy_classpath(retry=False)
        raise self.Error('Ivy bootstrapping failed - invalid classpath: %s' % ':'.join(cp))
      return cp

  def _bootstrap_ivy(self, bootstrap_jar_cache_path):
    if not os.path.exists(bootstrap_jar_cache_path):
      with temporary_file() as bootstrap_jar:
        fetcher = Fetcher()
        checksummer = fetcher.ChecksumListener(digest=hashlib.sha1())
        try:
          print('Downloading %s' % self._bootstrap_jar_url)
          fetcher.download(self._bootstrap_jar_url,
                           listener=fetcher.ProgressListener().wrap(checksummer),
                           path_or_fd=bootstrap_jar,
                           timeout=self._timeout)
          print('sha1: %s' % checksummer.checksum)
          bootstrap_jar.close()
          touch(bootstrap_jar_cache_path)
          shutil.move(bootstrap_jar.name, bootstrap_jar_cache_path)
        except fetcher.Error as e:
          raise self.Error('Problem fetching the ivy bootstrap jar! %s' % e)

    return Ivy(bootstrap_jar_cache_path,
               ivy_settings=self.ivy_settings,
               ivy_cache_dir=self.ivy_cache_dir)


class Ivy(object):
  """Encapsulates the ivy cli taking care of the basic invocation letting you just worry about the
  args to pass to the cli itself.
  """

  class Error(Exception):
    """Indicates an error executing an ivy command."""

  def __init__(self, classpath, java_executor=None, ivy_settings=None, ivy_cache_dir=None):
    """Configures an ivy wrapper for the ivy distribution at the given classpath."""

    self._classpath = maybe_list(classpath)

    self._java = java_executor or SubprocessExecutor()
    if not isinstance(self._java, Executor):
      raise ValueError('java_executor must be an Executor instance, given %s of type %s'
                       % (self._java, type(self._java)))

    self._ivy_settings = ivy_settings
    if self._ivy_settings and not isinstance(self._ivy_settings, Compatibility.string):
      raise ValueError('ivy_settings must be a string, given %s of type %s'
                       % (self._ivy_settings, type(self._ivy_settings)))

    self._ivy_cache_dir = ivy_cache_dir
    if self._ivy_cache_dir and not isinstance(self._ivy_cache_dir, Compatibility.string):
      raise ValueError('ivy_cache_dir must be a string, given %s of type %s'
                       % (self._ivy_cache_dir, type(self._ivy_cache_dir)))

  def execute(self, args=None, jvm_args=None):
    """Executes the ivy commandline client with the given args.

    Raises Ivy.Error if the command fails for any reason.
    """
    if self._ivy_cache_dir and '-cache' not in args:
      # TODO(John Sirois): Currently this is a magic property to support hand-crafted <caches/> in
      # ivysettings.xml.  Ideally we'd support either simple -caches or these hand-crafted cases
      # instead of just hand-crafted.  Clean this up by taking over ivysettings.xml and generating
      # it from BUILD constructs.
      jvm_args = ['-Divy.cache.dir=%s' % self._ivy_cache_dir] + (jvm_args or [])

    if self._ivy_settings and '-settings' not in args:
      args = ['-settings', self._ivy_settings] + args

    try:
      result = self._java.execute(self._classpath, 'org.apache.ivy.Main', args=args,
                                  jvm_args=jvm_args)
      if result != 0:
        raise self.Error('Ivy command failed with exit code %d%s'
                         % (result, ': ' + ' '.join(args) if args else ''))
    except self._java.Error as e:
      raise self.Error('Problem executing ivy: %s' % e)



# ==================================================================================================
# Copyright 2012 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

from collections import defaultdict

from twitter.pants import get_buildroot
from twitter.pants.tasks import Task


class CodeGen(Task):
  """
    Encapsulates the common machinery for codegen targets that support multiple output languages.
    This Task will only invoke code generation for changed targets and for the set of languages
    in the active context that require codegen unless forced.
  """

  def is_gentarget(self, target):
    """
      Subclasses must override and return True if the target is a target it handles generating
      code for.
    """
    raise NotImplementedError

  def is_forced(self, lang):
    """
      Subclasses should override and return True if they wish to force code generation for the
      given language.
    """
    return False

  def genlangs(self):
    """
      Subclasses must override and return a dict mapping supported generation target language names
      to a predicate that can select targets consuming that language.
    """
    raise NotImplementedError

  def prepare_gen(self, targets):
    """
      Subclasses should override if they need to prepare for potential upcoming calls to genlang.

      Note that this does not mean genlang will necessarily be called.
    """
    pass

  def genlang(self, lang, targets):
    """
      Subclasses must override and generate code in :lang for the given targets.
    """
    raise NotImplementedError

  def createtarget(self, lang, gentarget, dependees):
    """
      Subclasses must override and create a synthetic target containing the sources generated for
      the given :gentarget.
    """
    raise NotImplementedError

  def getdependencies(self, gentarget):
    # TODO(John Sirois): fix python/jvm dependencies handling to be uniform
    if hasattr(gentarget, 'internal_dependencies'):
      return gentarget.internal_dependencies
    else:
      return gentarget.dependencies

  def updatedependencies(self, target, dependency):
    if hasattr(target, 'update_dependencies'):
      target.update_dependencies([dependency])
    else:
      target.dependencies.add(dependency)

  def execute(self, targets):
    gentargets = [t for t in targets if self.is_gentarget(t)]
    capabilities = self.genlangs() # lang_name => predicate
    gentargets_by_dependee = self.context.dependants(
      on_predicate=self.is_gentarget,
      from_predicate=lambda t: not self.is_gentarget(t)
    )
    dependees_by_gentarget = defaultdict(set)
    for dependee, tgts in gentargets_by_dependee.items():
      for gentarget in tgts:
        dependees_by_gentarget[gentarget].add(dependee)

    def find_gentargets(predicate):
      tgts = set()
      for dependee in gentargets_by_dependee.keys():
        if predicate(dependee):
          for tgt in gentargets_by_dependee.pop(dependee):
            tgt.walk(tgts.add, self.is_gentarget)
      return tgts.intersection(set(gentargets))

    gentargets_bylang = {}
    for lang, predicate in capabilities.items():
      gentargets_bylang[lang] = gentargets if self.is_forced(lang) else find_gentargets(predicate)
    if gentargets_by_dependee:
      self.context.log.warn('Left with unexpected unconsumed gen targets:\n\t%s' % '\n\t'.join(
        '%s -> %s' % (dependee, gentargets)
        for dependee, gentargets in gentargets_by_dependee.items()
      ))

    if gentargets:
      self.prepare_gen(gentargets)
      with self.changed(gentargets, invalidate_dependants=True) as changed_targets:
        changed = set(changed_targets)
        for lang, tgts in gentargets_bylang.items():
          lang_changed = changed.intersection(tgts)
          if lang_changed:
            self.genlang(lang, lang_changed)

    # Link synthetic targets for all in-play gen targets
    for lang, tgts in gentargets_bylang.items():
      if tgts:
        langtarget_by_gentarget = {}
        for target in tgts:
          tgt = self.createtarget(
            lang,
            target,
            dependees_by_gentarget.get(target, [])
          )
          tgt.derived_from = target
          langtarget_by_gentarget[target] = tgt
        genmap = self.context.products.get(lang)
        for gentarget, langtarget in langtarget_by_gentarget.items():
          genmap.add(gentarget, get_buildroot(), [langtarget])
          for dep in self.getdependencies(gentarget):
            if self.is_gentarget(dep):
              self.updatedependencies(langtarget, langtarget_by_gentarget[dep])



# ==================================================================================================
# Copyright 2011 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

from __future__ import print_function

import os
import sys

from twitter.common.dirutil import safe_mkdir

from twitter.pants.ivy import Bootstrapper, Ivy
from twitter.pants.python.code_generator import CodeGenerator


class PythonAntlrBuilder(CodeGenerator):
  """
    Antlr builder.
  """
  def run_antlrs(self, output_dir):
    args = [
      '-dependency', 'org.antlr', 'antlr', self.target.antlr_version,
      '-types', 'jar',
      '-main', 'org.antlr.Tool',
      '--', '-fo', output_dir
    ]
    for source in self.target.sources:
      abs_path = os.path.abspath(os.path.join(self.root, self.target.target_base, source))
      args.append(abs_path)

    try:
      ivy = Bootstrapper.default_ivy()
      ivy.execute(args)
      return True
    except (Bootstrapper.Error, Ivy.Error) as e:
      print('ANTLR generation failed! %s' % e, file=sys.stderr)
      return False

  def generate(self):
    # Create the package structure.
    path = self.sdist_root

    package = ''
    for module_name in self.target.module.split('.'):
      path = os.path.join(path, module_name)
      if package == '':
        package = module_name
      else:
        package = package + '.' + module_name
      safe_mkdir(path)
      with open(os.path.join(path, '__init__.py'), 'w') as f:
        if package != self.target.module:  # Only write this in the non-leaf modules.
          f.write("__import__('pkg_resources').declare_namespace(__name__)")
          self.created_namespace_packages.add(package)
      self.created_packages.add(package)

    # autogenerate the python files that we bundle up
    self.run_antlrs(path)



# ==================================================================================================
# Copyright 2011 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

import os

from twitter.pants.base.target import Target


def resolve_target_sources(target_sources, extension=None, relative_to_target_base=False):
  """Given a list of pants targets, extract their sources as a list.

  Filters against the extension if given and optionally returns the paths relative to the target
  base.
  """
  resolved_sources = []
  for resolved in Target.resolve_all(target_sources):
    if hasattr(resolved, 'sources'):
      resolved_sources.extend(
        source if relative_to_target_base else os.path.join(resolved.target_base, source)
        for source in resolved.sources if not extension or source.endswith(extension)
      )
  return resolved_sources


from .annotation_processor import AnnotationProcessor
from .anonymous import AnonymousDeps
from .artifact import Artifact
from .benchmark import Benchmark
from .credentials import Credentials
from .doc import Page, Wiki
from .exclude import Exclude
from .exportable_jvm_library import ExportableJvmLibrary
from .gem import Gem
from .hadoop_binary import TwitterHadoopBinary
from .idl_jar_thrift_library import IdlJvmThriftLibrary
from .internal import InternalTarget
from .jar_dependency import JarDependency
from .jar_library import JarLibrary
from .java_antlr_library import JavaAntlrLibrary
from .java_library import JavaLibrary
from .java_thrift_library import JavaThriftLibrary
from .java_thriftstore_dml_library import JavaThriftstoreDMLLibrary
from .java_protobuf_library import JavaProtobufLibrary
from .java_tests import JavaTests
from .jvm_binary import Bundle, JvmApp, JvmBinary
from .jvm_target import JvmTarget
from .oink_query import OinkQuery
from .pants_target import Pants
from .python_artifact import PythonArtifact
from .python_binary import PythonBinary
from .python_egg import PythonEgg
from .python_library import PythonLibrary
from .python_antlr_library import PythonAntlrLibrary
from .python_thrift_library import PythonThriftLibrary
from .python_requirement import PythonRequirement
from .python_target import PythonTarget
from .python_tests import PythonTests, PythonTestSuite
from .repository import Repository
from .ruby_thrift_library import RubyThriftLibrary
from .ruby_target import RubyTarget
from .resources import Resources
from .scala_library import ScalaLibrary
from .scala_tests import ScalaTests
from .scalac_plugin import ScalacPlugin
from .sources import SourceRoot
from .storm_binary import TwitterStormBinary
from .thrift_library import ThriftJar, ThriftLibrary
from .with_sources import TargetWithSources

__all__ = [
  'AnnotationProcessor',
  'AnonymousDeps',
  'Artifact',
  'Benchmark',
  'Bundle',
  'Credentials',
  'Exclude',
  'ExportableJvmLibrary',
  'Gem',
  'IdlJvmThriftLibrary',
  'InternalTarget',
  'JarDependency',
  'JarLibrary',
  'JavaAntlrLibrary',
  'JavaLibrary',
  'JavaThriftLibrary',
  # TODO(Anand) Remove this from pants proper when a code adjoinment mechanism exists
  # or ok if/when thriftstore is open sourced as well..
  'JavaThriftstoreDMLLibrary',
  'JavaProtobufLibrary',
  'JavaTests',
  'JvmApp',
  'JvmBinary',
  'JvmTarget',
  'OinkQuery',
  'Page',
  'Pants',
  'PythonArtifact',
  'PythonBinary',
  'PythonEgg',
  'PythonLibrary',
  'PythonTarget',
  'PythonAntlrLibrary',
  'PythonRequirement',
  'PythonThriftLibrary',
  'PythonTests',
  'PythonTestSuite',
  'Repository',
  'Resources',
  'RubyTarget',
  'RubyThriftLibrary',
  'ScalaLibrary',
  'ScalaTests',
  'ScalacPlugin',
  'SourceRoot',
  'TargetWithSources',
  'ThriftJar',
  'ThriftLibrary',
  'TwitterHadoopBinary',
  'TwitterStormBinary',
  'Wiki'
]



# ==================================================================================================
# Copyright 2011 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

import os
import collections

from twitter.common.collections import OrderedSet, maybe_list
from twitter.common.lang import Compatibility

from twitter.pants import is_concrete

from .address import Address
from .build_manual import manual
from .parse_context import ParseContext


class TargetDefinitionException(Exception):
  """Thrown on errors in target definitions."""

  def __init__(self, target, msg):
    address = getattr(target, 'address', None)
    if address is None:
      try:
        location = ParseContext.locate().current_buildfile
      except ParseContext.ContextError:
        location = 'unknown location'
      address = 'unknown target of type %s in %s' % (target.__class__.__name__, location)
    super(Exception, self).__init__('Error with %s: %s' % (address, msg))


@manual.builddict()
class Target(object):
  """The baseclass for all pants targets.

  Handles registration of a target amongst all parsed targets as well as location of the target
  parse context.
  """

  _targets_by_address = None
  _addresses_by_buildfile = None

  @classmethod
  def get_all_addresses(cls, buildfile):
    """Returns all of the target addresses in the specified buildfile if already parsed; otherwise,
    parses the buildfile to find all the addresses it contains and then returns them.
    """
    def lookup():
      if buildfile in cls._addresses_by_buildfile:
        return cls._addresses_by_buildfile[buildfile]
      else:
        return OrderedSet()

    addresses = lookup()
    if addresses:
      return addresses
    else:
      ParseContext(buildfile).parse()
      return lookup()

  @classmethod
  def _clear_all_addresses(cls):
    cls._targets_by_address = {}
    cls._addresses_by_buildfile = collections.defaultdict(OrderedSet)

  @classmethod
  def get(cls, address):
    """Returns the specified module target if already parsed; otherwise, parses the buildfile in the
    context of its parent directory and returns the parsed target.
    """
    def lookup():
      return cls._targets_by_address[address] if address in cls._targets_by_address else None

    target = lookup()
    if target:
      return target
    else:
      ParseContext(address.buildfile).parse()
      return lookup()

  @classmethod
  def resolve_all(cls, targets, *expected_types):
    """Yield the resolved concrete targets checking each is a subclass of one of the expected types
    if specified.
    """
    if targets:
      for target in maybe_list(targets, expected_type=Target):
        for resolved in filter(is_concrete, target.resolve()):
          if expected_types and not isinstance(resolved, expected_types):
            raise TypeError('%s requires types: %s and found %s' % (cls, expected_types, resolved))
          yield resolved

  def __init__(self, name):
    """
    :param string name: The target name.
    """
    if not isinstance(name, Compatibility.string):
      self.address = '%s:%s' % (ParseContext.locate().current_buildfile, str(name))
      raise TargetDefinitionException(self, "Invalid target name: %s" % name)
    self.name = name
    self.is_codegen = False
    self.description = None

    self.address = self._locate()

    # TODO(John Sirois): Transition all references to self.identifier to eliminate id builtin
    # ambiguity
    self.id = self._create_id()

    self._register()

    # For synthetic codegen targets this will be the original target from which
    # the target was synthesized.
    self.derived_from = self

  def _post_construct(self, func, *args, **kwargs):
    """Registers a command to invoke after this target's BUILD file is parsed."""
    ParseContext.locate().on_context_exit(func, *args, **kwargs)

  def _create_id(self):
    """Generates a unique identifier for the BUILD target.

    The generated id is safe for use as a path name on unix systems.
    """
    buildfile_relpath = os.path.dirname(self.address.buildfile.relpath)
    if buildfile_relpath in ('.', ''):
      return self.name
    else:
      return "%s.%s" % (buildfile_relpath.replace(os.sep, '.'), self.name)

  def _locate(self):
    parse_context = ParseContext.locate()
    return Address(parse_context.current_buildfile, self.name)

  def _register(self):
    existing = self._targets_by_address.get(self.address)
    if existing and existing is not self:
      if existing.address.buildfile != self.address.buildfile:
        raise TargetDefinitionException(self, "already defined in a sibling BUILD "
                                              "file: %s" % existing.address.buildfile.relpath)
      else:
        raise TargetDefinitionException(self, "duplicate to %s" % existing)

    self._targets_by_address[self.address] = self
    self._addresses_by_buildfile[self.address.buildfile].add(self.address)

  @property
  def identifier(self):
    """A unique identifier for the BUILD target.

    The generated id is safe for use as a path name on unix systems.
    """
    return self.id

  def resolve(self):
    """Yields an iterator over all concrete targets this target represents."""
    yield self

  def walk(self, work, predicate=None):
    """Walk of this target's dependency graph visiting each node exactly once.

    If a predicate is supplied it will be used to test each target before handing the target to
    work and descending. Work can return targets in which case these will be added to the walk
    candidate set if not already walked.

    :param work: Callable that takes a :py:class:`twitter.pants.base.target.Target`
      as its single argument.
    :param predicate: Callable that takes a :py:class:`twitter.pants.base.target.Target`
      as its single argument and returns True if the target should passed to ``work``.
    """
    if not callable(work):
      raise ValueError('work must be callable but was %s' % work)
    if predicate and not callable(predicate):
      raise ValueError('predicate must be callable but was %s' % predicate)
    self._walk(set(), work, predicate)

  def _walk(self, walked, work, predicate=None):
    for target in self.resolve():
      if target not in walked:
        walked.add(target)
        if not predicate or predicate(target):
          additional_targets = work(target)
          if hasattr(target, '_walk'):
            target._walk(walked, work, predicate)
          if additional_targets:
            for additional_target in additional_targets:
              if hasattr(additional_target, '_walk'):
                additional_target._walk(walked, work, predicate)

  @manual.builddict()
  def with_description(self, description):
    """Set a human-readable description of this target."""
    self.description = description
    return self

  def __eq__(self, other):
    return isinstance(other, Target) and self.address == other.address

  def __hash__(self):
    return hash(self.address)

  def __ne__(self, other):
    return not self.__eq__(other)

  def __repr__(self):
    return "%s(%s)" % (type(self).__name__, self.address)

Target._clear_all_addresses()




# DO NOT EDIT THIS FILE -- AUTOGENERATED BY PANTS
# Target: PythonLibrary(src/python/twitter/pants/BUILD:pants-packaged)

from setuptools import setup

setup(**
{   'classifiers': [   'Intended Audience :: Developers',
                       'License :: OSI Approved :: Apache Software License',
                       'Operating System :: OS Independent',
                       'Programming Language :: Python'],
    'description': "Twitter's pants build tool.",
    'entry_points': {   'console_scripts': [   'pants = twitter.pants.bin.pants_exe:main']},
    'install_requires': [   'twitter.common.quantity==0.1.0',
                            'twitter.common.python==0.3.1',
                            'twitter.common.util==0.1.0',
                            'twitter.common.config==0.1.0',
                            'pytest==2.3.5',
                            'Mako==0.4.0',
                            'twitter.common.decorators==0.1.0',
                            'Markdown==2.1.1',
                            'twitter.common.process==0.1.0',
                            'psutil==1.1.2',
                            'twitter.common.confluence==0.1.0',
                            'requests==2.0.0',
                            'elementtree==1.2.7-20070827-preview',
                            'Pygments==1.4',
                            'ansicolors==1.0.2',
                            'pytest-cov==1.6'],
    'license': 'Apache License, Version 2.0',
    'name': 'twitter.pants',
    'namespace_packages': ['twitter', 'twitter.common'],
    'package_data': {   'twitter.pants': [   'docs/targets_graph.dot',
                                             'docs/invoking.rst',
                                             'docs/howto_contribute.rst',
                                             'docs/pagerank.py',
                                             'docs/howto_develop.rst',
                                             'docs/3rdparty_jvm.rst',
                                             'docs/first_tutorial.rst',
                                             'GROUPS',
                                             'docs/first_concepts.rst',
                                             'docs/docs.rst',
                                             'docs/publish_via_git.sh',
                                             'docs/gen.py',
                                             'docs/tasks_graph_tsort_specs.dot',
                                             'docs/Makefile',
                                             'docs/conf.py',
                                             'docs/pants-logo.png',
                                             'docs/taskdev/falcon.gif',
                                             'docs/README.rst',
                                             'docs/taskdev/taskdev.asc',
                                             'docs/target_addresses.rst',
                                             'docs/pants-logo.ico',
                                             'docs/dev_tasks.rst',
                                             'docs/3rdparty_py.rst',
                                             'docs/targets_graph_gen.dot',
                                             'docs/migration.rst',
                                             'docs/publish.rst',
                                             'docs/build_files.rst',
                                             'docs/setup_repo.rst',
                                             'BUILD',
                                             'docs/3rdparty.rst',
                                             'README.md',
                                             'docs/_templates/layout.html',
                                             'OWNERS',
                                             'docs/ThriftDeps.rst',
                                             'docs/tasks_graph.dot',
                                             'docs/internals.rst',
                                             'docs/pants_essentials_techtalk.asc',
                                             'docs/from_maven.rst',
                                             'docs/credits.rst',
                                             'docs/python-readme.rst',
                                             'docs/install.rst',
                                             'docs/tasks_graph_tsort.dot',
                                             'docs/JVMProjects.rst',
                                             'docs/index.rst'],
                        'twitter.pants.java': ['bin/ng.py'],
                        'twitter.pants.java.distribution': [   'Version.java',
                                                               'README.md',
                                                               'Version.class'],
                        'twitter.pants.python': [   'BUILD.twitter',
                                                    'BUILD',
                                                    'README.md'],
                        'twitter.pants.tasks': [   'templates/eclipse/pydevproject-3.7.mk',
                                                   'templates/builddictionary/goals_reference.mk',
                                                   'templates/jar_publish/pom.mk',
                                                   'templates/idea/module-12.mk',
                                                   'templates/eclipse/debug-launcher-3.7.mk',
                                                   'files/eclipse/org.eclipse.core.resources.prefs',
                                                   'files/eclipse/org.eclipse.jdt.ui.prefs',
                                                   'templates/ivy_resolve/ivy.mk',
                                                   'templates/eclipse/project-3.7.mk',
                                                   'templates/jar_publish/ivysettings.mk',
                                                   'templates/eclipse/org.eclipse.jdt.core.prefs-3.7.mk',
                                                   'templates/eclipse/classpath-3.7.mk',
                                                   'templates/idea/project-12.mk',
                                                   'templates/builddictionary/page.mk',
                                                   'templates/eclipse/factorypath-3.7.mk'],
                        'twitter.pants.tasks.javadoc.assets': [   'javadoc.html',
                                                                  'favicon.ico',
                                                                  'index.html',
                                                                  'javadoc-single-package.html'],
                        'twitter.pants.version': ['OWNERS']},
    'package_dir': {   '': 'src'},
    'packages': [   'twitter',
                    'twitter.common',
                    'twitter.common.python',
                    'twitter.common.python.bin',
                    'twitter.pants',
                    'twitter.pants.base',
                    'twitter.pants.bin',
                    'twitter.pants.buildtimestats',
                    'twitter.pants.commands',
                    'twitter.pants.engine',
                    'twitter.pants.fs',
                    'twitter.pants.goal',
                    'twitter.pants.ivy',
                    'twitter.pants.java',
                    'twitter.pants.java.distribution',
                    'twitter.pants.java.jar',
                    'twitter.pants.net',
                    'twitter.pants.net.http',
                    'twitter.pants.process',
                    'twitter.pants.python',
                    'twitter.pants.scm',
                    'twitter.pants.targets',
                    'twitter.pants.tasks',
                    'twitter.pants.tasks.javadoc',
                    'twitter.pants.tasks.javadoc.assets',
                    'twitter.pants.tasks.python',
                    'twitter.pants.version',
                    'twitter.thrift',
                    'twitter.thrift.text'],
    'url': 'https://github.com/twitter/commons',
    'version': '0.0.14-dev0',
    'zip_safe': True}
)



# ==================================================================================================
# Copyright 2012 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

__author__ = 'John Sirois'

try:
  import markdown

  WIKILINKS_PATTERN = r'\[\[([^\]]+)\]\]'

  class WikilinksPattern(markdown.inlinepatterns.Pattern):
    def __init__(self, build_url, markdown_instance=None):
      markdown.inlinepatterns.Pattern.__init__(self, WIKILINKS_PATTERN, markdown_instance)
      self.build_url = build_url

    def handleMatch(self, m):
      alias, url = self.build_url(m.group(2).strip())
      el = markdown.util.etree.Element('a')
      el.set('href', url)
      el.text = markdown.util.AtomicString(alias)
      return el

  class WikilinksExtension(markdown.Extension):
    def __init__(self, build_url, configs=None):
      markdown.Extension.__init__(self, configs or {})
      self.build_url = build_url

    def extendMarkdown(self, md, md_globals):
      md.inlinePatterns['wikilinks'] = WikilinksPattern(self.build_url, md)

  HAS_MARKDOWN = True
except ImportError:
  HAS_MARKDOWN = False

try:
  from pygments.formatters.html import HtmlFormatter
  from pygments.styles import get_all_styles

  def configure_codehighlight_options(option_group, mkflag):
    all_styles = list(get_all_styles())
    option_group.add_option(mkflag("code-style"), dest="markdown_to_html_code_style",
                            type="choice", choices=all_styles,
                            help="Selects the stylesheet to use for code highlights, one of: "
                                 "%s." % ' '.join(all_styles))

  def emit_codehighlight_css(path, style):
    with safe_open(path, 'w') as css:
      css.write((HtmlFormatter(style=style)).get_style_defs('.codehilite'))
    return path
except ImportError:
  def configure_codehighlight_options(option_group, mkflag): pass
  def emit_codehighlight_css(path, style): pass


import codecs
import os
import re
import textwrap

from twitter.common.dirutil import safe_mkdir, safe_open

from twitter.pants import binary_util, get_buildroot
from twitter.pants.base import Address, Target
from twitter.pants.targets import Page
from twitter.pants.tasks import Task, TaskError

class MarkdownToHtml(Task):
  AVAILABLE = HAS_MARKDOWN

  @classmethod
  def setup_parser(cls, option_group, args, mkflag):
    configure_codehighlight_options(option_group, mkflag)

    option_group.add_option(mkflag("open"), mkflag("open", negate=True),
                            dest = "markdown_to_html_open",
                            action="callback", callback=mkflag.set_bool, default=False,
                            help = "[%default] Open the generated documents in a browser.")

    option_group.add_option(mkflag("fragment"), mkflag("fragment", negate=True),
                            dest = "markdown_to_html_fragment",
                            action="callback", callback=mkflag.set_bool, default=False,
                            help = "[%default] Generate a fragment of html to embed in a page.")

    option_group.add_option(mkflag("outdir"), dest="markdown_to_html_outdir",
                            help="Emit generated html in to this directory.")

    option_group.add_option(mkflag("extension"), dest = "markdown_to_html_extensions",
                            action="append",
                            help = "Override the default markdown extensions and process pages "
                                   "whose source have these extensions instead.")

  def __init__(self, context):
    Task.__init__(self, context)

    self.open = context.options.markdown_to_html_open

    pants_workdir = context.config.getdefault('pants_workdir')
    self.outdir = (
      context.options.markdown_to_html_outdir
      or context.config.get('markdown-to-html',
                            'workdir',
                            default=os.path.join(pants_workdir, 'markdown'))
    )

    self.extensions = set(
      context.options.markdown_to_html_extensions
      or context.config.getlist('markdown-to-html', 'extensions', default=['.md', '.markdown'])
    )

    self.fragment = context.options.markdown_to_html_fragment

    self.code_style = context.config.get('markdown-to-html', 'code-style', default='friendly')
    if hasattr(context.options, 'markdown_to_html_code_style'):
      if context.options.markdown_to_html_code_style:
        self.code_style = context.options.markdown_to_html_code_style

  def execute(self, targets):
    if not MarkdownToHtml.AVAILABLE:
      raise TaskError('Cannot process markdown - no markdown lib on the sys.path')

    # TODO(John Sirois): consider adding change detection

    css_relpath = os.path.join('css', 'codehighlight.css')
    css = emit_codehighlight_css(os.path.join(self.outdir, css_relpath), self.code_style)
    if css:
      self.context.log.info('Emitted %s' % css)

    def is_page(target):
      return isinstance(target, Page)

    roots = set()
    interior_nodes = set()
    if self.open:
      dependencies_by_page = self.context.dependants(on_predicate=is_page, from_predicate=is_page)
      roots.update(dependencies_by_page.keys())
      for dependencies in dependencies_by_page.values():
        interior_nodes.update(dependencies)
        roots.difference_update(dependencies)
      for page in self.context.targets(is_page):
        # There are no in or out edges so we need to show show this isolated page.
        if not page.dependencies and page not in interior_nodes:
          roots.add(page)

    plaingenmap = self.context.products.get('markdown_html')
    wikigenmap = self.context.products.get('wiki_html')
    show = []
    for page in filter(is_page, targets):
      _, ext = os.path.splitext(page.source)
      if ext in self.extensions:
        def process_page(key, outdir, url_builder, config, genmap, fragment=False):
          html_path = self.process(
            outdir,
            page.target_base,
            page.source,
            self.fragment or fragment,
            url_builder,
            config,
            css=css
          )
          self.context.log.info('Processed %s to %s' % (page.source, html_path))
          relpath = os.path.relpath(html_path, outdir)
          genmap.add(key, outdir, [relpath])
          return html_path

        def url_builder(linked_page, config=None):
          path, ext = os.path.splitext(linked_page.source)
          return linked_page.name, os.path.relpath(path + '.html', os.path.dirname(page.source))

        page_path = os.path.join(self.outdir, 'html')
        html = process_page(page, page_path, url_builder, lambda p: None, plaingenmap)
        if css and not self.fragment:
          plaingenmap.add(page, self.outdir, list(css_relpath))
        if self.open and page in roots:
          show.append(html)

        for wiki in page.wikis():
          def get_config(page):
            return page.wiki_config(wiki)
          basedir = os.path.join(self.outdir, wiki.id)
          process_page((wiki, page), basedir, wiki.url_builder, get_config,
                       wikigenmap, fragment=True)

    if show:
      binary_util.ui_open(*show)

  PANTS_LINK = re.compile(r'''pants\(['"]([^)]+)['"]\)(#.*)?''')

  def process(self, outdir, base, source, fragmented, url_builder, get_config, css=None):
    def parse_url(spec):
      match = MarkdownToHtml.PANTS_LINK.match(spec)
      if match:
        page = Target.get(Address.parse(get_buildroot(), match.group(1)))
        anchor = match.group(2) or ''
        if not page:
          raise TaskError('Invalid link %s' % match.group(1))
        alias, url = url_builder(page, config=get_config(page))
        return alias, url + anchor
      else:
        return spec, spec

    def build_url(label):
      components = label.split('|', 1)
      if len(components) == 1:
        return parse_url(label.strip())
      else:
        alias, link = components
        _, url = parse_url(link.strip())
        return alias, url

    wikilinks = WikilinksExtension(build_url)

    path, ext = os.path.splitext(source)
    output_path = os.path.join(outdir, path + '.html')
    safe_mkdir(os.path.dirname(output_path))
    with codecs.open(output_path, 'w', 'utf-8') as output:
      with codecs.open(os.path.join(get_buildroot(), base, source), 'r', 'utf-8') as input:
        md_html = markdown.markdown(
          input.read(),
          extensions=['codehilite(guess_lang=False)', 'extra', 'tables', 'toc', wikilinks],
        )
        if fragmented:
          if css:
            with safe_open(css) as fd:
              output.write(textwrap.dedent('''
              <style type="text/css">
              %s
              </style>
              ''').strip() % fd.read())
              output.write('\n')
          output.write(md_html)
        else:
          if css:
            css_relpath = os.path.relpath(css, outdir)
            out_relpath = os.path.dirname(source)
            link_relpath = os.path.relpath(css_relpath, out_relpath)
            css = '<link rel="stylesheet" type="text/css" href="%s"/>' % link_relpath
          html = textwrap.dedent('''
          <html>
            <head>
              <meta charset="utf-8">
              %s
            </head>
            <body>
          <!-- generated by pants! -->
          %s
            </body>
          </html>
          ''').strip() % (css or '', md_html)
          output.write(html)
        return output.name



# ==================================================================================================
# Copyright 2012 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

__author__ = 'John Sirois'

import os
import re

from collections import defaultdict

from twitter.pants import get_buildroot

class Dependencies(object):
  """
    Can map source files to class files by parsing jvm compiler dependency files with lines of the
    form:

    [source file path] -> [class file path]

    All paths are assumed to be normalized to be relative to the classfile output directory.
  """

  _CLASS_FILE_NAME_PARSER = re.compile(r'(?:\$.*)*\.class$')

  def __init__(self, outputdir, depfile):
    self.outputdir = outputdir
    self.depfile = depfile

  def findclasses(self, targets):
    """
      Returns a mapping from a target to its source to classes mapping.
      For example:

      dependencies = Dependencies(outdir, depfile)
      mapping = dependencies.findclasses(targets)
      for target, src_to_classes in mapping.items():
        for source, classes in src_to_classes.items():
          print('source: %s produces classes: %s' % (
            os.path.join(target.target_base, source),
            [os.path.join(outdir, cls) for cls in classes]
          ))
    """

    sources = set()
    target_by_source = dict()
    for target in targets:
      for source in target.sources:
        src = os.path.normpath(os.path.join(target.target_base, source))
        target_by_source[src] = target
        sources.add(src)

    classes_by_target_by_source = defaultdict(lambda: defaultdict(set))
    if os.path.exists(self.depfile):
      with open(self.depfile, 'r') as deps:
        for dep in deps.readlines():
          src, cls = dep.strip().split('->')
          sourcefile = os.path.relpath(os.path.join(self.outputdir, src.strip()), get_buildroot())
          if sourcefile in sources:
            classfile = os.path.relpath(os.path.join(self.outputdir, cls.strip()), self.outputdir)
            target = target_by_source[sourcefile]
            relsrc = os.path.relpath(sourcefile, target.target_base)
            classes_by_target_by_source[target][relsrc].add(classfile)
    return classes_by_target_by_source



# ==================================================================================================
# Copyright 2013 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

from __future__ import print_function

import os
import shutil
import tempfile
import textwrap

from twitter.common.dirutil.chroot import RelativeChroot
from twitter.pants.python.sdist_builder import SdistBuilder


class CodeGenerator(object):
  class Error(Exception): pass
  class CodeGenerationException(Error): pass

  def __init__(self, target, root_dir, config, target_suffix=None):
    self.target = target
    self.config = config
    self.suffix = target_suffix or ''
    self.root = root_dir
    distdir = self.config.getdefault('pants_distdir')
    self.chroot = RelativeChroot(root_dir, distdir, target.name)
    codegen_root = tempfile.mkdtemp(dir=self.chroot.path(), prefix='codegen.')
    self.codegen_root = os.path.relpath(codegen_root, self.chroot.path())
    self.created_packages = set()
    self.created_namespace_packages = set()

  def __del__(self):
    self.cleanup()

  def cleanup(self):
    shutil.rmtree(self.chroot.path())

  @staticmethod
  def path_to_module(path):
    return path.replace(os.path.sep, '.')

  def package_name(self):
    return '%s%s' % (self.target.id, self.suffix)

  def requirement_string(self):
    return '%s==0.0.0' % self.package_name()

  @property
  def package_dir(self):
    """Return the code generation root."""
    return "."

  def generate(self):
    """Generate code for this target, updating the sets .created_packages and
       .created_namespace_packages."""
    raise NotImplementedError

  def dump_setup_py(self):
    boilerplate = textwrap.dedent("""
      from setuptools import setup

      setup(name        = "%(package_name)s",
            version     = "0.0.0",
            description = "autogenerated code for %(target_name)s",
            package_dir = { "": %(package_dir)r },
            packages    = %(packages)s,
            namespace_packages = %(namespace_packages)s)
    """)
    boilerplate = boilerplate % {
      'target_name': self.target.id,
      'package_name': self.package_name(),
      'package_dir': self.package_dir,
      'target_name': self.target.name,
      'packages': repr(self.created_packages),
      'namespace_packages': repr(list(self.created_namespace_packages))
    }
    self.chroot.write(boilerplate.encode('utf8'), os.path.join(self.codegen_root, 'setup.py'))
    self.chroot.write("include *.py".encode('utf8'), os.path.join(self.codegen_root, 'MANIFEST.in'))

  @property
  def sdist_root(self):
    return os.path.join(self.chroot.path(), self.codegen_root)

  @property
  def package_root(self):
    return os.path.join(self.sdist_root, self.package_dir)

  def build(self, interpreter=None):
    self.generate()
    self.dump_setup_py()
    return SdistBuilder.build(self.sdist_root, self.target, interpreter=interpreter)



# ==================================================================================================
# Copyright 2012 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

import re

from itertools import izip_longest


class Revision(object):
  """Represents a software revision that is comparable to another revision describing the same
  software.
  """
  class BadRevision(Exception):
    """Indicates a problem parsing a revision."""

  @classmethod
  def _parse_atom(cls, atom):
    try:
      return int(atom)
    except ValueError:
      return atom

  @classmethod
  def semver(cls, rev):
    """Attempts to parse a Revision from a semantic version.

    See http://semver.org/ for the full specification.
    """
    def parse_extra(delimiter, value):
      if not value:
        return None, None
      else:
        components = value.split(delimiter, 1)
        return components[0], None if len(components) == 1 else components[1]

    def parse_patch(patch):
      patch, pre_release = parse_extra('-', patch)
      if pre_release:
        pre_release, build = parse_extra('+', pre_release)
      else:
        patch, build = parse_extra('+', patch)
      return patch, pre_release, build

    def parse_components(value):
      if not value:
        yield None
      else:
        for atom in value.split('.'):
          yield cls._parse_atom(atom)

    try:
      major, minor, patch = rev.split('.', 2)
      patch, pre_release, build = parse_patch(patch)
      components = [int(major), int(minor), int(patch)]
      components.extend(parse_components(pre_release))
      components.extend(parse_components(build))
      return cls(*components)
    except ValueError:
      raise cls.BadRevision("Failed to parse '%s' as a semantic version number" % rev)

  @classmethod
  def lenient(cls, rev):
    """A lenient revision parser that tries to split the version into logical components with
    heuristics inspired by PHP's version_compare.
    """
    rev = re.sub(r'(\d)([a-zA-Z])', r'\1.\2', rev)
    rev = re.sub(r'([a-zA-Z])(\d)', r'\1.\2', rev)
    return cls(*map(cls._parse_atom, re.split(r'[.+_\-]', rev)))

  def __init__(self, *components):
    self._components = components

  @property
  def components(self):
    """Returns a list of this revision's components from most major to most minor."""
    return list(self._components)

  def __cmp__(self, other):
    for ours, theirs in izip_longest(self._components, other._components, fillvalue=0):
      difference = cmp(ours, theirs)
      if difference != 0:
        return difference
    return 0

  def __repr__(self):
    return '%s(%s)' % (self.__class__.__name__, ', '.join(map(repr, self._components)))



# ==================================================================================================
# Copyright 2013 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

from __future__ import print_function

import os
import re
import sys

from twitter.common import app
from twitter.common.log import LogOptions

from twitter.pants.java import NailgunClient


# TODO(John Sirois): Extract this to a file a release toolchain can edit.
VERSION = '0.0.2'


app.add_option('--nailgun-version',
               dest='just_version',
               default=False,
               action='store_true',
               help='print product version and exit')

app.add_option('--nailgun-showversion',
               dest='show_version',
               default=False,
               action='store_true',
               help='print product version and continue')

app.add_option('--nailgun-server',
               dest='ng_host',
               metavar='HOST',
               default=NailgunClient.DEFAULT_NG_HOST,
               help='to specify the address of the nailgun server (default is %default)')

app.add_option('--nailgun-port',
               dest='ng_port',
               metavar='PORT',
               default=NailgunClient.DEFAULT_NG_PORT,
               type='int',
               help='to specify the port of the nailgun server (default is %default)')

app.add_option('--nailgun-help',
               dest='show_help',
               default=False,
               action='store_true',
               help='print this message and exit')

app.set_usage('''%(command)s class [--nailgun-options] [args]
        (to execute a class)
 or: %(command)s alias [options] [args]
        (to execute an aliased class)
 or: alias [options] [args]
        (to execute an aliased class, where "alias"
         is both the alias for the class and a symbolic
         link to the ng client)
''' % dict(command=sys.argv[0]))


def main(args):
  options = app.get_options()
  if options.show_help:
    app.help()

  if options.show_version or options.just_version:
    print('Python NailGun client version %s' % VERSION)
    if options.just_version:
      sys.exit(0)

  # Assume ng.pex has been aliased to the command name
  command = re.compile('.pex$').sub('', os.path.basename(sys.argv[0]))
  args_index = 0

  # Otherwise the command name is the 1st arg
  if command == 'ng':
    if not args:
      app.help()

    command = args[0]
    args_index = 1

  ng = NailgunClient(host=options.ng_host, port=options.ng_port)
  try:
    result = ng(command, *args[args_index:], **os.environ)
    sys.exit(result)
  except ng.NailgunError as e:
    print('Problem executing command: %s' % e, file=sys.stderr)
    sys.exit(1)


LogOptions.disable_disk_logging()
app.main()



# ==================================================================================================
# Copyright 2012 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

import os

from collections import defaultdict

from twitter.common.collections import OrderedSet
from twitter.pants import get_buildroot
from twitter.pants.base import manual, ParseContext


@manual.builddict()
class SourceRoot(object):
  """Allows registration of a source root for a set of targets.

  A source root is the base path sources for a particular language are found relative to.
  Generally compilers or interpreters for the source will expect sources relative to a base path
  and a source root allows calculation of the correct relative paths.

  E.g., a Java compiler probably expects to find ``.java`` files for
  ``package com.twitter.common.net`` in ``*something*/com/twitter/common/net``.
  The ``source_root`` command specifies that *something*.

  It is illegal to have nested source roots.
  """
  _ROOTS_BY_TYPE = defaultdict(OrderedSet)
  _TYPES_BY_ROOT = defaultdict(OrderedSet)
  _SEARCHED = set()

  @classmethod
  def reset(cls):
    """Reset all source roots to empty. Only intended for testing."""
    cls._ROOTS_BY_TYPE = defaultdict(OrderedSet)
    cls._TYPES_BY_ROOT = defaultdict(OrderedSet)
    cls._SEARCHED = set()

  @staticmethod
  def _register(sourceroot):
    for t in sourceroot.types:
      SourceRoot._ROOTS_BY_TYPE[t].add(sourceroot.basedir)
      SourceRoot._TYPES_BY_ROOT[sourceroot.basedir].add(t)

  @staticmethod
  def find(target):
    """Finds the source root for the given target.  If none is registered, the parent
    directory of the target's BUILD file is returned.
    """
    target_path = os.path.relpath(target.address.buildfile.parent_path, get_buildroot())

    def _find():
      for typ in target.__class__.mro():
        for root in SourceRoot._ROOTS_BY_TYPE.get(typ, ()):
          if target_path.startswith(root):
            return root

    # Try already registered roots
    root = _find()
    if root:
      return root

    # Fall back to searching the ancestor path for a root
    for buildfile in reversed(target.address.buildfile.ancestors()):
      if buildfile not in SourceRoot._SEARCHED:
        SourceRoot._SEARCHED.add(buildfile)
        ParseContext(buildfile).parse()
        root = _find()
        if root:
          return root

    # Finally, resolve files relative to the BUILD file parent dir as the target base
    return target_path

  @staticmethod
  def types(root):
    """Returns the set of target types rooted at root."""
    return SourceRoot._TYPES_BY_ROOT[root]

  @staticmethod
  def roots(target_type):
    """Returns the set of roots for given target type."""
    return SourceRoot._ROOTS_BY_TYPE[target_type]

  @staticmethod
  def all_roots():
    """Returns a mapping from source roots to the associated target types."""
    return dict(SourceRoot._TYPES_BY_ROOT)

  @staticmethod
  def here(*types):
    """Registers the cwd as a source root for the given target types."""
    return SourceRoot.register(None, *types)

  @staticmethod
  def register(basedir, *types):
    """Registers the given basedir as a source root for the given target types."""
    return SourceRoot(basedir, *types)

  def __init__(self, basedir, *types):
    """Initializes a source root at basedir for the given target types.

    :param string basedir: The base directory to resolve sources relative to
    :param types: The target types to register ``basedir`` as a source root for
    """
    basepath = os.path.abspath(basedir or os.path.curdir)
    if get_buildroot() != os.path.commonprefix((basepath, get_buildroot())):
      raise ValueError('The supplied basedir %s is not a sub-path of the project root %s' % (
        basepath,
        get_buildroot()
      ))

    self.basedir = os.path.relpath(basepath, get_buildroot())
    self.types = types
    SourceRoot._register(self)



# ==================================================================================================
# Copyright 2013 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

import os.path
import re

from twitter.pants import has_sources
from twitter.pants.base.config import Config
from twitter.pants.process.xargs import Xargs
from twitter.pants.tasks import TaskError
from twitter.pants.tasks.nailgun_task import NailgunTask


class Scalastyle(NailgunTask):
  """Checks scala source files to ensure they're stylish.

  Scalastyle is configured via the 'scalastyle' pants.ini section.

  * ``config`` - Required path of the scalastyle configuration file.
  * ``excludes`` - Optional path of an excludes file that contains
    lines of regular expressions used to exclude matching files
    from style checks. File names matched against these regular
    expressions are relative to the repository root
    (e.g.: com/twitter/mybird/MyBird.scala).
  """

  _CONFIG_SECTION = 'scalastyle'
  _MAIN = 'org.scalastyle.Main'

  @classmethod
  def setup_parser(cls, option_group, args, mkflag):
    NailgunTask.setup_parser(option_group, args, mkflag)

    option_group.add_option(mkflag("skip"), mkflag("skip", negate=True),
                            dest="scalastyle_skip", default=False,
                            action="callback", callback=mkflag.set_bool,
                            help="[%default] Skip scalastyle.")

  def __init__(self, context):
    NailgunTask.__init__(self, context)
    self._scalastyle_config = self.context.config.get_required(
      Scalastyle._CONFIG_SECTION, 'config')
    if not os.path.exists(self._scalastyle_config):
      raise Config.ConfigError(
        'Scalastyle config file does not exist: %s' % self._scalastyle_config)

    excludes_file = self.context.config.get(Scalastyle._CONFIG_SECTION, 'excludes')
    self._excludes = set()
    if excludes_file:
      if not os.path.exists(excludes_file):
        raise Config.ConfigError(
          'Scalastyle excludes file does not exist: %s' % excludes_file)
      self.context.log.debug('Using scalastyle excludes file %s' % excludes_file)
      with open(excludes_file) as fh:
        for pattern in fh.readlines():
          self._excludes.add(re.compile(pattern.strip()))

  def execute(self, targets):
    if self.context.options.scalastyle_skip:
      self.context.log.debug('Skipping checkstyle.')
      return

    check_targets = list()
    for target in targets:
      for tgt in target.resolve():
        if has_sources(tgt, '.scala'):
          check_targets.append(tgt)

    def filter_excludes(filename):
      if self._excludes:
        for exclude in self._excludes:
          if exclude.match(filename):
            return False
      return True

    scala_sources = list()
    for target in check_targets:
      def collect(filename):
        if filename.endswith('.scala'):
          scala_sources.append(os.path.join(target.target_base, filename))
      map(collect, filter(filter_excludes, target.sources))

    if scala_sources:
      def call(srcs):
        return self.runjava(main=Scalastyle._MAIN,
                            classpath=self.profile_classpath('scalastyle'),
                            args=['-c', self._scalastyle_config] + srcs)
      result = Xargs(call).execute(scala_sources)
      if result != 0:
        raise TaskError('java %s ... exited non-zero (%i)' % (Scalastyle._MAIN, result))



# ==================================================================================================
# Copyright 2012 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

import itertools
import os

from twitter.pants.targets import TargetWithSources
from twitter.pants.targets.jvm_binary import JvmApp
from twitter.pants.tasks.console_task import ConsoleTask

__author__ = 'Dave Buchfuhrer'

class FileDeps(ConsoleTask):
  def console_output(self, targets):
    files = set()
    for target in targets:
      if isinstance(target, TargetWithSources):
        files.update(target.expand_files(recursive=False))
      if isinstance(target, JvmApp):
        files.update(itertools.chain(*[bundle.filemap.keys() for bundle in target.bundles]))
    return files



# ==================================================================================================
# Copyright 2012 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================


from __future__ import print_function

__author__ = 'Tejal Desai'

import httplib
import json
import os
import socket
import sys
import urllib
import time

from twitter.common import dirutil, log
from twitter.common.contextutil import temporary_file
from twitter.common.dirutil import safe_open, safe_mkdtemp
from twitter.common.dirutil.fileset import Fileset
from twitter.common.quantity import Amount, Time
from twitter.common.quantity.parse_simple import parse_time, InvalidTime
from twitter.common.util.command_util import CommandUtil


MAX_RECORDS = 100
LOG_FILES = {
              "DEBUG" : ".global.DEBUG",
              "ERROR" : ".global.ERROR",
              "FATAL" : ".global.FATAL",
              "INFO" : ".global.INFO",
              "WARNING" : ".global.WARNING"
            }

class StatsHttpClient(object):
  def __init__(self, host=None, port=None, http_endpoint=None, stats_dir=None):
    self._client = None
    self._host = host
    self._port = port
    self._http_endpoint = http_endpoint
    self._stats_dir = stats_dir

  def _get_client(self):
    if self._client is None:
      self._client = httplib.HTTPConnection(self._host, self._port, timeout=2)
    return self._client

  def push_stats(self, json_stats):
    log.debug("Uploading pants stats to %s" % self._host)
    client = self._get_client()
    params = urllib.urlencode({"json": json_stats})
    headers = {"Content-type": "application/x-www-form-urlencoded", "Accept": "text/plain"}
    client.request("POST", self._http_endpoint, params, headers)
    resp = client.getresponse()
    resp_type = resp.status / 100
    client.close()
    if resp_type not in [2, 3]:
      log.debug("There was an error uploading the stats")

  def process_stats_file(self):
    for filename in Fileset.walk(self._stats_dir):
      try :
        with open(os.path.join(self._stats_dir, filename), 'r') as stats_file:
          lines = stats_file.readlines()
          tmp_str = ",".join(lines)
          tmp_str.strip(',')
          self.push_stats("[" + tmp_str + "]")
        os.remove(os.path.join(self._stats_dir, filename))
      except httplib.HTTPException as e:
        log.debug("HTTPException %s" % e)
      except OSError as e:
        log.debug("Error reading or deleting a stats file %s" % e)


class StatsUploader():
  def __init__(self, host, port, endpoint, max_delay, stats_file, user, force_stats_upload=False):
    self.force_stats_upload = force_stats_upload
    self._stats_log_dir = dirutil.safe_mkdtemp()
    self._stats_log_file = os.path.join(self._stats_log_dir, "current_run")
    log.init(self._stats_log_file)
    self._stats_dir = os.path.join("/tmp", user, "stats_uploader_dir")
    self._stats_http_client = StatsHttpClient(host, port, endpoint, self._stats_dir)
    self._max_delay = max_delay
    self._pants_stat_file = stats_file
    self._user = user

  def upload_sync(self, stats):
    try:
      last_modified = self.collect_host_env_info(stats)
      if not last_modified:
        last_modified = int(os.path.getmtime(self._pants_stat_file))

      with safe_open(self._pants_stat_file, 'r') as stats_file:
        lines = stats_file.readlines()
      #Just want to make sure, we do not wait for MAX_RECORDS but also upload when
      #the last time we uploaded is less than configured value in the pants.ini
      last_uploaded = Amount(int(time.time()) - last_modified, Time.SECONDS)
      if (self.force_stats_upload or len(lines) >= MAX_RECORDS or last_uploaded > self._max_delay):
        #Put the file in the right place.
        dirutil.safe_mkdir(self._stats_dir)
        with temporary_file(self._stats_dir, False) as stats_uploader_tmpfile:
          os.rename(self._pants_stat_file, stats_uploader_tmpfile.name)
        self._stats_http_client.process_stats_file()
      #Merge Logs so that user /tmp is not cluttered with too many log files for each run.
      self.merge_logs()
      sys.exit(0)
    except OSError as e:
      log.debug("Error manipulating stats files for upload %s" % e)

  def collect_host_env_info(self,stats):
    #Get Environment Variable
    stats["env"] = os.environ.data
    stats["timestamp"] = int(time.time())
    try:
      #Get the System info
      import psutil
      stats["cpu_time"] = psutil.cpu_percent(interval=1)
      stats["network_counter"] = psutil.net_io_counters()
      stats["no_of_cpus"] = psutil.NUM_CPUS
    except Exception as e:
      log.debug("Exception %s. Cannot collect psutil stats" % e)

    #Get Git info
    stats["git"] = {}
    (ret, git_origin) = CommandUtil().execute_and_get_output(["git", "remote", "-v"])
    if ret == 0:
      for url in git_origin.splitlines():
        origin = url.split()
        str1 = origin[2].strip("(").strip(")")
        if origin:
          stats["git"][str1] = origin[1]

    #Get git branch
    (ret, git_branch) = CommandUtil().execute_and_get_output(["git", "rev-parse", "--abbrev-ref",
                                                              "HEAD"])
    if ret == 0:
      stats["git"]["branch"] = git_branch.strip()
    #Network IP
    try:
      stats["ip"] = socket.gethostbyname(socket.gethostname())
    except Exception as e:
      log.debug("Exception %s. Cannot get ip stats" % e)
    log.debug("Done collecting stats")

    #get the last modified time for the File so that we can upload the stats if they havent being
    #Uploaded for last 6 hours.
    last_modified = None
    if os.path.exists(self._pants_stat_file):
      last_modified = int(os.path.getmtime(self._pants_stat_file))
    try:
      with open(self._pants_stat_file , 'a') as stats_file:
        json_response = json.dumps(stats, cls=PythonObjectEncoder)
        stats_file.write(json_response + "\n")
      return last_modified
    except IOError as e:
      log.debug("Could not write the pants stats %s" % e)

  def merge_logs(self):
    prefix = os.path.join("/tmp", self._user, "buildtime_uploader")
    self.append_log_file("DEBUG")
    self.append_log_file("INFO")
    self.append_log_file("FATAL")
    self.append_log_file("ERROR")
    self.append_log_file("WARNING")

  def append_log_file(self, log_level):
    target_file = os.path.join("/tmp", self._user, "buildtime_uploader" + LOG_FILES[log_level])
    source_file = self._stats_log_file + "." + log_level
    self._append_to_file(source_file, target_file)

  def _append_to_file(self, source_file, target_file):
    with open(target_file, 'a') as target_fh:
      with open(source_file, "r") as source_fh:
        target_fh.write(source_fh.read())

class PythonObjectEncoder(json.JSONEncoder):
  def default(self, obj):
    try:
     return json.JSONEncoder.default(self, obj)
    except TypeError as e:
      log.warn("Could not encode %s" % obj)
      pass



# ==================================================================================================
# Copyright 2011 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

from twitter.common.collections import OrderedSet
from twitter.common.python.interpreter import PythonIdentity
from twitter.pants.base.target import TargetDefinitionException
from twitter.pants.targets.with_dependencies import TargetWithDependencies
from twitter.pants.targets.with_sources import TargetWithSources


class PythonTarget(TargetWithDependencies, TargetWithSources):
  """Base class for all Python targets."""

  def __init__(self,
               name,
               sources,
               resources=None,
               dependencies=None,
               provides=None,
               compatibility=None):
    TargetWithSources.__init__(self, name, sources=sources)
    TargetWithDependencies.__init__(self, name, dependencies=dependencies)
    self.resources = self._resolve_paths(self.target_base, resources) if resources else OrderedSet()
    self.provides = provides
    self.compatibility = compatibility or ['']
    for req in self.compatibility:
      try:
        PythonIdentity.parse_requirement(req)
      except ValueError as e:
        raise TargetDefinitionException(str(e))




# ==================================================================================================
# Copyright 2011 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

import os

from ..base import BuildFile, Target
from .. import get_buildroot

from .console_task import ConsoleTask


class Filemap(ConsoleTask):
  """Outputs a mapping from source file to the target that owns the source file."""

  def console_output(self, _):
    visited = set()
    for target in self._find_targets():
      if target not in visited:
        visited.add(target)
        if hasattr(target, 'sources') and target.sources is not None:
          for sourcefile in target.sources:
            path = os.path.join(target.target_base, sourcefile)
            yield '%s %s' % (path, target.address)

  def _find_targets(self):
    if len(self.context.target_roots) > 0:
      for target in self.context.target_roots:
        yield target
    else:
      for buildfile in BuildFile.scan_buildfiles(get_buildroot()):
        target_addresses = Target.get_all_addresses(buildfile)
        for target_address in target_addresses:
          yield Target.get(target_address)



# =============================================================================
# Copyright 2013 Twitter, Inc.
# -----------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

import cgi
import inspect
import os

import twitter.pants
import twitter.pants.base

from twitter.common.dirutil import Fileset, safe_open
from twitter.pants.base.generator import Generator, TemplateData
from twitter.pants.goal.phase import Phase
from twitter.pants.tasks import Task, TaskError

from pkg_resources import resource_string


def entry(nom, classdoc=None, msg_rst=None, argspec=None, funcdoc=None, methods=None):
  """Create a struct that our template expects to see.

  :param nom: Symbol name, e.g. python_binary
  :param classdoc: plain text appears above argspec
  :param msg_rst: reST. useful in hand-crafted entries
  :param argspec: arg string like (x, y="deflt")
  :param funcdoc: function's __doc__, plain text
  :param methods: list of entries for class' methods
  """

  def indent_docstring_by_1(s):
    """Given a non-empty docstring, return a version indented by a space.
    Given an empty thing, return the thing itself
    """
    # In reST, it's useful to have strings that are similarly-indented.
    # If we have a classdoc indented by 2 next to an __init__ funcdoc indented
    # by 4, reST doesn't format things nicely. Oh, totally-dedenting doesn't
    # format nicely either.

    # Docstring indentation: more gnarly than you'd think:
    # http://www.python.org/dev/peps/pep-0257/#handling-docstring-indentation
    if not s: return s
    # Convert tabs to spaces (following the normal Python rules)
    # and split into a list of lines:
    lines = s.expandtabs().splitlines()
    # Determine minimum indentation (first line doesn't count):
    indent = 999
    for line in lines[1:]:
        stripped = line.lstrip()
        if stripped:
            indent = min(indent, len(line) - len(stripped))
    # Remove indentation (first line is special):
    trimmed = [lines[0].strip()]
    if indent < 999:
        for line in lines[1:]:
            trimmed.append(line[indent:].rstrip())
    # Strip off trailing and leading blank lines:
    while trimmed and not trimmed[-1]:
        trimmed.pop()
    while trimmed and not trimmed[0]:
        trimmed.pop(0)
    # Return a single string:
    return '\n'.join([" " + t for t in trimmed])

  return TemplateData(
    nom=nom.strip(),
    classdoc=indent_docstring_by_1(classdoc),
    msg_rst=indent_docstring_by_1(msg_rst),
    argspec=argspec,
    funcdoc=indent_docstring_by_1(funcdoc),
    methods=methods)


def msg_entry(nom, defn):
  """For hard-wired entries a la "See Instead" or other simple stuff"""
  return entry(nom, msg_rst=defn)


def entry_for_one_func(nom, func):
  """Generate a BUILD dictionary entry for a function
  nom: name like 'python_binary'
  func: function object"""
  args, varargs, varkw, defaults = inspect.getargspec(func)
  argspec = inspect.formatargspec(args, varargs, varkw, defaults)
  return entry(nom,
               argspec=argspec,
               funcdoc=func.__doc__)

def entry_for_one_method(nom, method):
  """Generate a BUILD dictionary entry for a method
  nom: name like 'with_description'
  method: method object"""
  # TODO(lhosken) : This is darned similar to entry_for_one_func. Merge 'em?
  #                 (Punted so far since funcdoc indentation made my head hurt)
  assert inspect.ismethod(method)
  args, varargs, varkw, defaults = inspect.getargspec(method)
  # args[:1] instead of args to discard "self" arg
  argspec = inspect.formatargspec(args[1:], varargs, varkw, defaults)
  return entry(nom,
               argspec=argspec,
               funcdoc=(method.__doc__ or "").replace("\n", " "))


def entry_for_one(nom, sym):
  if inspect.isclass(sym):
    return entry_for_one_class(nom, sym)
  if inspect.ismethod(sym) or inspect.isfunction(sym):
    return entry_for_one_func(nom, sym)
  return msg_entry(nom, cgi.escape("TODO! no doc gen for %s %s" % (
        str(type(sym)), str(sym))))


PREDEFS = {  # some hardwired entries
  "Amount": {"defn": msg_entry("Amount", """
                                `Amount from twitter.commons.quantity <https://github.com/twitter/commons/blob/master/src/python/twitter/common/quantity/__init__.py>`_
                                E.g., ``Amount(2, Time.MINUTES)``.""")},
  "__builtins__": {"suppress": True},
  "Config": {"suppress": True},
  "Context": {"suppress": True},
  "__file__": {"defn": msg_entry("__file__", "Path to BUILD file (string).")},
  "get_buildroot": {"defn": msg_entry("get_buildroot",
                    """Alternate way to get `ROOT_DIR`_""")},
  "globs": {"defn": entry_for_one("globs", Fileset.globs)},
  "group": {"suppress": True},
  "jar_library": {"defn": msg_entry("jar_library",
                  """Old name for `dependencies`_""")},
  "java_tests": {"defn": msg_entry("java_tests",
                  """Old name for `junit_tests`_""")},
  "JavaLibrary": {"suppress": True},
  "JavaTests": {"suppress": True},
  "maven_layout": {"defn": entry_for_one("maven_layout",
                                          twitter.pants.maven_layout)},
  "oink_query": {"suppress": True},
  "python_artifact": {"suppress": True},
  "rglobs": {"defn": entry_for_one("rglobs", Fileset.rglobs)},
  "ROOT_DIR": {"defn": msg_entry("ROOT_DIR",
                                  "Root directory of source code (string).")},
  "scala_tests": {"defn": msg_entry("scala_tests",
                  """Old name for `scala_specs`_""")},
  "Time": {"defn": msg_entry("Time", """
                             `Amount from twitter.commons.quantity <https://github.com/twitter/commons/blob/master/src/python/twitter/common/quantity/__init__.py>`_
                             E.g., ``Amount(2, Time.MINUTES)``."""), },
}


# Thingies like scala_library
# Returns list of duples [(name, object), (name, object), (name, object),...]
def get_syms():
  r = {}
  vc = twitter.pants.base.ParseContext.default_globals()
  for s in vc:
    if s in PREDEFS: continue
    o = vc[s]
    r[s] = o
  return r


def tocl(d):
  """Generate TOC, in-page links to the IDs we're going to define below"""
  anchors = sorted(d.keys(), key=str.lower)
  return TemplateData(t="All The Things", e=[a for a in anchors])


def tags_tocl(d, tag_list, title):
  """Generate specialized TOC.
  E.g., tags_tocl(d, ["python", "anylang"], "Python")
  tag_list: if an entry's tags contains any of these, use it
  title: pretty title
  """
  filtered_anchors = []
  for anc in sorted(d.keys(), key=str.lower):
    entry = d[anc]
    if not "tags" in entry: continue
    found = [t for t in tag_list if t in entry["tags"]]
    if not found: continue
    filtered_anchors.append(anc)
  return TemplateData(t=title, e=filtered_anchors)


def entry_for_one_class(nom, klas):
  """  Generate a BUILD dictionary entry for a class.
  nom: name like 'python_binary'
  klas: class like twitter.pants.python_binary"""
  try:
    args, varargs, varkw, defaults = inspect.getargspec(klas.__init__)
    argspec = inspect.formatargspec(args[1:], varargs, varkw, defaults)
    funcdoc = klas.__init__.__doc__

    methods = []
    for attrname in dir(klas):
      attr = getattr(klas, attrname)
      attr_bdi = twitter.pants.base.get_builddict_info(attr)
      if not attr_bdi: continue
      if inspect.ismethod(attr):
        methods.append(entry_for_one_method(attrname, attr))
        continue
      raise TaskError('@manual.builddict on non-method %s within class %s but I only know what to do with methods' % (attrname, nom))

  except TypeError:  # __init__ might not be a Python function
    argspec = None
    funcdoc = None
    methods = None

  return entry(nom,
               classdoc=klas.__doc__,
               argspec=argspec,
               funcdoc=funcdoc,
               methods=methods)


def assemble(predefs=PREDEFS, symbol_hash=None):
  """Assemble big hash of entries suitable for smushing into a template.

  predefs: Hash of "hard-wired" predefined entries.
  symbol_hash: Python syms from which to generate more entries. Default: get from BUILD context"""
  d = {}
  for k in PREDEFS:
    v = PREDEFS[k]
    if "suppress" in v and v["suppress"]: continue
    d[k] = v
  if symbol_hash is None:
    symbol_hash = get_syms()
  for k in symbol_hash:
    bdi = twitter.pants.base.get_builddict_info(symbol_hash[k])
    if bdi is None: continue
    d[k] = bdi.copy()
    if not "defn" in d[k]:
      d[k]["defn"] = entry_for_one(k, symbol_hash[k])
  return d


class BuildBuildDictionary(Task):
  """Generate documentation for the Sphinx site."""

  def __init__(self, context):
    Task.__init__(self, context)
    self._templates_dir = os.path.join('templates', 'builddictionary')
    self._outdir = os.path.join(self.context.config.getdefault("pants_distdir"), "builddict")


  def execute(self, targets):
    self._gen_goals_reference()

    d = assemble()
    template = resource_string(__name__, os.path.join(self._templates_dir, 'page.mk'))
    tocs = [tocl(d),
            tags_tocl(d, ["java", "scala", "jvm", "anylang"], "JVM"),
            tags_tocl(d, ["python", "anylang"], "Python")]
    defns = [d[t]["defn"] for t in sorted(d.keys(), key=str.lower)]
    filename = os.path.join(self._outdir, 'build_dictionary.rst')
    self.context.log.info('Generating %s' % filename)
    with safe_open(filename, 'w') as outfile:
      generator = Generator(template,
                            tocs=tocs,
                            defns=defns)
      generator.write(outfile)

  def _gen_goals_reference(self):
    """Generate the goals reference rst doc."""
    phases = dict()
    for phase, goals in Phase.all():
      phases[phase] = goals

    template = resource_string(__name__, os.path.join(self._templates_dir, 'goals_reference.mk'))
    filename = os.path.join(self._outdir, 'goals_reference.rst')
    self.context.log.info('Generating %s' % filename)
    with safe_open(filename, 'w') as outfile:
      generator = Generator(template, phases=phases)
      generator.write(outfile)



# ==================================================================================================
# Copyright 2011 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

from twitter.pants.base import manual

from .java_thrift_library import JavaThriftLibrary


# TODO(Anand): Remove this from pants proper when a code adjoinment mechanism exists
# or ok if/when thriftstore is open sourced as well..
@manual.builddict(tags=["java"])
class JavaThriftstoreDMLLibrary(JavaThriftLibrary):
  """Defines a target that builds java stubs from a thriftstore DDL file."""

  def __init__(self,
               name,
               sources,
               provides=None,
               dependencies=None):
    """
    :param string name: The name of this target, which combined with this
      build file defines the target :class:`twitter.pants.base.address.Address`.
    :param sources: A list of filenames representing the source code
      this library is compiled from.
    :type sources: list of strings
    :param Artifact provides:
      The :class:`twitter.pants.targets.artifact.Artifact`
      to publish that represents this target outside the repo.
    :param dependencies: List of :class:`twitter.pants.base.target.Target` instances
      this target depends on.
    :type dependencies: list of targets
    """

    JavaThriftLibrary.__init__(self,
                               name,
                               sources,
                               provides,
                               compiler='scrooge',
                               language='java',
                               dependencies=dependencies)
